cdef extern from "<numbers>" namespace "std::numbers" nogil:
    # C++20 mathematical constants
    const double e
    const double log2e
    const double log10e
    const double pi
    const double inv_pi
    const double inv_sqrtpi
    const double ln2
    const double ln10
    const double sqrt2
    const double sqrt3
    const double inv_sqrt3
    const double egamma
    const double phi
